#!/usr/bin/env python

"""
Evaluates a trained model by predicting new synthetic images and calculating the image similarity and the angle error
"""

import glob
import logging
import multiprocessing as mp
import os
import pickle
import random
import tempfile
from argparse import Namespace
from functools import partial
from typing import List, Optional

import numpy as np
import tensorflow as tf

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.config.default_paths import CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset import Dataset
from wormpose.dataset.loader import load_dataset, get_dataset_name
from wormpose.dataset.loaders.resizer import ResizeOptions
from wormpose.images.scoring import ResultsScoring, BaseScoringDataManager
from wormpose.machine_learning.best_models_saver import BestModels
from wormpose.machine_learning.generic_file_writer import GenericFileWriter
from wormpose.machine_learning.synthetic_data_generator import SyntheticDataGenerator
from wormpose.pose.distance_metrics import angle_distance
from wormpose.pose.eigenworms import load_eigenworms_matrix, theta_to_modes
from wormpose.pose.postures_model import PosturesModel
from wormpose.pose.results_datatypes import ShuffledResults

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _write_detailed_labeled_data_to_pickle(f, **kwargs):
    pickle.dump(
        (
            kwargs["image_data"],
            kwargs["template_measurements"],
            kwargs["template_frame"],
            kwargs["template_skeleton"],
            kwargs["theta"],
        ),
        f,
    )


class _PickleDetailedLabeledDataWriter(GenericFileWriter):
    """This labeled data writer saves detailed information about the synthetic image generation"""

    def __init__(self, filename):
        super().__init__(
            open_file=partial(open, filename, "wb"),
            write_file=lambda f, data: _write_detailed_labeled_data_to_pickle(f, **data),
        )


def _eval_data_gen(filenames: List[str]):
    for filename in filenames:
        with open(filename, "rb") as f:
            while True:
                try:
                    res = pickle.load(f)
                    im = res[0]
                    im = im[:, :, np.newaxis]
                    im = im.astype(np.float32) / 255
                    yield im
                except EOFError:
                    break


def _load_templates(pkl_filenames: List[str]):
    all_templates_data = []
    all_labels = []
    for pkl_filename in pkl_filenames:
        with open(pkl_filename, "rb") as pkl_file:
            try:
                while True:
                    (frame, template_measurements, template_frame, template_skeleton, label_theta) = pickle.load(
                        pkl_file
                    )
                    all_templates_data.append([template_frame, template_skeleton, template_measurements, frame])
                    all_labels.append(label_theta)
            except EOFError:
                pass
    return all_templates_data, all_labels


class _ScoringDataManager(BaseScoringDataManager):
    def __init__(self, pkl_filenames):
        self._all_templates_data, _ = _load_templates(pkl_filenames)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __getitem__(self, frame_index):
        return self._all_templates_data[frame_index]

    def __len__(self):
        return len(self._all_templates_data)


def _parse_arguments(dataset_path: str, kwargs: dict):
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("num_process") is None:
        kwargs["num_process"] = os.cpu_count()
    if kwargs.get("temp_dir") is None:
        kwargs["temp_dir"] = tempfile.gettempdir()
    if kwargs.get("batch_size") is None:
        kwargs["batch_size"] = 512
    if kwargs.get("num_samples") is None:
        kwargs["num_samples"] = 1000
    if kwargs.get("postures_generation") is None:
        kwargs["postures_generation"] = PosturesModel().generate
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None
    if kwargs.get("model_path") is None:
        kwargs["model_path"] = None
    if kwargs.get("random_seed") is None:
        kwargs["random_seed"] = None
    if kwargs.get("eigenworms_matrix_path") is None:
        kwargs["eigenworms_matrix_path"] = None
    kwargs["temp_dir"] = tempfile.mkdtemp(dir=kwargs["temp_dir"])

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("model_path") is None:
        default_models_dir = os.path.join(kwargs["experiment_dir"], default_paths.MODELS_DIRS)
        kwargs["model_path"] = BestModels(default_models_dir).best_model_path
    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def evaluate(dataset_path: str, **kwargs):
    """
    Evaluate a trained model by predicting synthetic data and recording the image similarity

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    mp.set_start_method("spawn", force=True)

    random.seed(args.random_seed)
    np.random.seed(args.random_seed)

    results_dir = os.path.join(args.experiment_dir, "evaluation")
    os.makedirs(results_dir, exist_ok=True)

    config = load_config(args.config)
    eigenworms_matrix = load_eigenworms_matrix(args.eigenworms_matrix_path)

    dataset = load_dataset(
        dataset_loader=config.dataset_loader,
        dataset_path=dataset_path,
        selected_video_names=args.video_names,
        resize_options=ResizeOptions(resize_factor=config.resize_factor),
    )

    pkl_filenames = _generate_synthetic_data(
        dataset, args.num_process, args.num_samples, args.postures_generation, args.temp_dir, args.random_seed,
    )

    keras_model = tf.keras.models.load_model(args.model_path, compile=False)

    tf_dataset = tf.data.Dataset.from_generator(
        partial(_eval_data_gen, pkl_filenames), tf.float32, tf.TensorShape(dataset.image_shape + (1,)),
    ).batch(args.batch_size)

    network_predictions = keras_model.predict(tf_dataset)[: args.num_samples]
    shuffled_results = ShuffledResults(random_theta=network_predictions)

    ResultsScoring(
        frame_preprocessing=dataset.frame_preprocessing,
        num_process=args.num_process,
        temp_dir=args.temp_dir,
        image_shape=dataset.image_shape,
    )(
        results=shuffled_results, scoring_data_manager=_ScoringDataManager(pkl_filenames),
    )
    # Keep the maximum score between the two head/tail options for this evaluation
    image_scores = np.max(shuffled_results.scores, axis=1)

    # Now calculate the angle error and mode error
    angle_error = []
    modes_error = []
    theta_predictions = []
    _, theta_labels = _load_templates(pkl_filenames)
    for theta_label, theta_results in zip(theta_labels, shuffled_results.theta):
        dists = [angle_distance(theta_result, theta_label) for theta_result in theta_results]
        closest_index = int(np.argmin(dists))
        closest_theta = theta_results[closest_index]
        theta_predictions.append(closest_theta)
        angle_error.append(dists[closest_index])
        if eigenworms_matrix is not None:
            modes_label = theta_to_modes(theta_label, eigenworms_matrix)
            modes_prediction = theta_to_modes(closest_theta, eigenworms_matrix)
            mode_error = np.abs(modes_label - modes_prediction)
            modes_error.append(mode_error)

    np.savetxt(os.path.join(results_dir, "image_score.txt"), image_scores)
    np.savetxt(os.path.join(results_dir, "angle_error.txt"), angle_error)
    np.savetxt(os.path.join(results_dir, "theta_labels.txt"), theta_labels)
    np.savetxt(os.path.join(results_dir, "theta_predictions.txt"), theta_predictions)
    if eigenworms_matrix is not None:
        np.savetxt(os.path.join(results_dir, "modes_error.txt"), modes_error)

    logger.info(
        f"Evaluated model with synthetic data,"
        f" average image similarity: {np.mean(image_scores):.4f},"
        f" average angle error (degrees): {np.rad2deg(np.mean(angle_error)):.2f}"
    )


def _generate_synthetic_data(
    dataset: Dataset,
    num_process: int,
    num_samples: int,
    postures_generation,
    temp_dir: str,
    random_seed: Optional[int],
):
    syn_data_file_pattern = os.path.join(temp_dir, "synthetic_{index}.pkl")
    synthetic_data_generator = SyntheticDataGenerator(
        num_process=num_process,
        temp_dir=temp_dir,
        dataset=dataset,
        postures_generation_fn=postures_generation,
        writer=_PickleDetailedLabeledDataWriter,
        enable_random_augmentations=False,
        random_seed=random_seed,
    )
    gen = synthetic_data_generator.generate(num_samples=num_samples, file_pattern=syn_data_file_pattern)
    list(gen)
    pkl_filenames = list(sorted(glob.glob(syn_data_file_pattern.format(index="*"))))
    return pkl_filenames


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_path", type=str)
    parser.add_argument("--model_path", type=str, help="Load models from this path.")
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only evaluate using a subset of videos. " "If not set, will include all videos in dataset_path.",
    )
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument(
        "--num_samples", type=int, help="How many synthetic samples to evaluate the model with",
    )
    parser.add_argument(
        "--eigenworms_matrix_path", help="Path to optional eigenworms matrix to also calculate mode error",
    )
    add_config_argument(parser)
    parser.add_argument("--temp_dir", type=str, help="Where to store temporary intermediate results")
    parser.add_argument("--num_process", type=int, help="How many worker processes")
    parser.add_argument("--batch_size", type=int)
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")
    args = parser.parse_args()

    evaluate(**vars(args))


if __name__ == "__main__":
    main()



"""
Generates evaluation data: random real processed images with labels and save them to a Tfrecord file
"""

import csv
import logging
import os

import numpy as np

from wormpose.dataset import Dataset
from wormpose.images.real_dataset import RealDataset
from wormpose.machine_learning import tfrecord_file
from wormpose.pose.centerline import skeleton_to_angle, flip_theta

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def generate(dataset: Dataset, num_samples: int, theta_dims: int, file_pattern: str,) -> int:
    """
    Generates evaluation dataset composed of processed real images, saved to a .TFrecord

    :param dataset: WormPose Dataset
    :param num_samples: How many images to generate
    :param theta_dims: Dimensions of theta for the labels
    :param file_pattern: Path of the output files with "index" variable
        example: "path_to_out/eval_{index}.tfrecord"
    :return: How many samples where actually generated (if there is less data than the requested num_samples)
    """
    labelled_frames = {}
    for video_name in dataset.video_names:
        skel_is_not_nan = ~np.any(np.isnan(dataset.features_dataset[video_name].skeletons), axis=(1, 2))
        labelled_indexes = np.where(skel_is_not_nan)[0]
        if len(labelled_indexes) > 0:
            labelled_frames[video_name] = labelled_indexes

    if len(labelled_frames) == 0:
        raise RuntimeError("Can't create evaluation data because couldn't find any labelled frame in the dataset.")

    len_labelled_frames = int(np.sum([len(x) for x in labelled_frames.values()]))
    if len_labelled_frames < num_samples:
        logging.warning(
            f"Not enough labelled frames in the dataset "
            f"to create an evaluation set of {num_samples} unique samples, "
            f"using all available {len_labelled_frames} samples instead."
        )
        num_samples = len_labelled_frames

    real_dataset = RealDataset(dataset.frame_preprocessing, dataset.image_shape)

    tfrecord_filename = file_pattern.format(index=0)
    csv_infos_filename = os.path.splitext(tfrecord_filename)[0] + ".csv"

    # get num_samples total random labelled frames from the videos
    eval_frames = _populate_eval_frames(labelled_frames, num_samples)

    # write the eval.tfrecord file with the images and the labels, save also the source infos in a separate eval.csv
    # the frames are not shuffled by video, all the frames from one video are consecutive in the file
    with tfrecord_file.Writer(tfrecord_filename) as record_writer, open(csv_infos_filename, "w") as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=" ", quotechar="|", quoting=csv.QUOTE_MINIMAL)
        for video_name, cur_video_eval_indexes in eval_frames.items():

            with dataset.frames_dataset.open(video_name) as frames:

                for eval_frame_index in cur_video_eval_indexes:
                    image_data, _ = real_dataset.process_frame(frames[eval_frame_index])
                    cur_skel = dataset.features_dataset[video_name].skeletons[eval_frame_index]
                    cur_theta = skeleton_to_angle(cur_skel, theta_dims=theta_dims)
                    cur_theta_flipped = flip_theta(cur_theta)

                    record_writer.write(image_data, cur_theta, cur_theta_flipped)
                    csv_writer.writerow([video_name, int(eval_frame_index)])

    return num_samples


def _populate_eval_frames(labelled_frames: dict, num_samples: int):
    cur_index = 0
    eval_frames = {}

    while cur_index < num_samples:
        # pick randomly a video name and a valid frame index for that video, without repetition
        chosen_video_name = np.random.choice(list(labelled_frames.keys()))
        valid_index_in_video = labelled_frames[chosen_video_name]
        pick_index_at = np.random.randint(0, len(valid_index_in_video))
        chosen_frame_index = valid_index_in_video[pick_index_at]
        # remove chosen frame from the available frames
        labelled_frames[chosen_video_name] = np.delete(labelled_frames[chosen_video_name], pick_index_at)
        if len(labelled_frames[chosen_video_name]) == 0:
            del labelled_frames[chosen_video_name]

        if chosen_video_name in eval_frames:
            eval_frames[chosen_video_name].append(chosen_frame_index)
        else:
            eval_frames[chosen_video_name] = [chosen_frame_index]
        cur_index += 1

    # sort for faster read access
    for video_name in eval_frames:
        eval_frames[video_name].sort()

    return eval_frames



"""
This module performs image scoring on the results (shuffled: two scores per frame or not)
It distributes the work to several processes
"""

import logging
import os
import pickle
from multiprocessing import Process, Manager
from typing import Tuple

import numpy as np

from wormpose import BaseFramePreprocessing
from wormpose.images.scoring.scoring_data_manager import BaseScoringDataManager
from wormpose.images.scoring.centerline_accuracy_check import CenterlineAccuracyCheck
from wormpose.pose.results_datatypes import BaseResults

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

_CHUNK_SIZE = 100


class ResultsScoring(object):
    def __init__(
        self,
        frame_preprocessing: BaseFramePreprocessing,
        image_shape: Tuple[int, int],
        temp_dir: str,
        num_process: int,
    ):
        self.frame_preprocessing = frame_preprocessing
        self.image_shape = image_shape
        self.temp_dir = temp_dir
        self.num_process = num_process
        self.data_reading_queue = None

    def __call__(
        self, results: BaseResults, scoring_data_manager: BaseScoringDataManager,
    ):

        manager = Manager()
        self.data_reading_queue = manager.Queue()
        results_queue = manager.Queue()

        workers = [
            Process(
                target=_compare_pred_real,
                args=(
                    self.frame_preprocessing,
                    self.data_reading_queue,
                    results_queue,
                    self.image_shape,
                    self.temp_dir,
                ),
            )
            for _ in range(self.num_process)
        ]
        for w in workers:
            w.start()

        self._produce_data_chunks(
            results_theta=results.theta, scoring_data_manager=scoring_data_manager,
        )

        for w in workers:
            w.join()

        results_files = []
        while not results_queue.empty():
            results_files.append(results_queue.get())
        results_files = sorted(results_files)

        matching_scores = []
        matching_skeletons = []
        for result_filename in results_files:
            with open(result_filename, "rb") as f:
                results_scores, results_skel = pickle.load(f)
            matching_scores.append(results_scores)
            matching_skeletons.append(results_skel)
            os.remove(result_filename)

        results.scores = np.concatenate(matching_scores)
        results.skeletons = np.concatenate(matching_skeletons)

    def _produce_data_chunks(self, results_theta: np.ndarray, scoring_data_manager: BaseScoringDataManager):

        chunk_data = []
        chunk_frames = []
        chunk_index = 0
        chunk_template_frames = []

        with scoring_data_manager as scoring_data:

            for frame_index, thetas in enumerate(results_theta):

                (template_frame, template_skeleton, template_measurements, frame,) = scoring_data[frame_index]

                chunk_template_frames.append(template_frame)
                chunk_data.append((thetas, template_skeleton, template_measurements))
                chunk_frames.append(frame)

                if len(chunk_data) >= _CHUNK_SIZE or frame_index == len(results_theta) - 1:
                    self._save_chunk(
                        chunk_data=chunk_data,
                        chunk_index=chunk_index,
                        real_frames=chunk_frames,
                        template_frames=chunk_template_frames,
                    )
                    chunk_frames = []
                    chunk_index += 1
                    chunk_data = []
                    chunk_template_frames = []

        for _ in range(self.num_process):
            self.data_reading_queue.put(None)

    def _save_chunk(self, chunk_data, chunk_index, real_frames, template_frames):
        compare_data_filename = os.path.join(self.temp_dir, f"compare_data_{chunk_index}.pkl")
        with open(compare_data_filename, "wb") as f:
            pickle.dump((chunk_data, real_frames, template_frames), f)

        self.data_reading_queue.put((compare_data_filename, chunk_index))


def _compare_pred_real(
    frame_preprocessing: BaseFramePreprocessing, data_reading_queue, results_queue, image_shape, temp_dir: str,
):
    centerline_accuracy = CenterlineAccuracyCheck(frame_preprocessing=frame_preprocessing, image_shape=image_shape)
    while True:
        queue_data = data_reading_queue.get()
        if queue_data is None:
            break

        compare_data_filename, chunk_index = queue_data

        with open(compare_data_filename, "rb") as f:
            chunk_data, chunk_real_frames, chunk_template_frames = pickle.load(f)
        os.remove(compare_data_filename)

        all_scores = None
        all_skel = None
        for index, (data, real_frame_orig, template_frame) in enumerate(
            zip(chunk_data, chunk_real_frames, chunk_template_frames)
        ):
            thetas, template_skeleton, template_measurements = data

            # reshape thetas so that we can handle all cases (one centerline or two or more centerlines to score)
            if len(thetas.shape) == 1:
                thetas = np.reshape(thetas, (1,) + thetas.shape)

            if all_scores is None:
                all_scores = np.empty((len(chunk_data), thetas.shape[0]), dtype=float)
            if all_skel is None:
                all_skel = np.empty(
                    (len(chunk_data), thetas.shape[0],) + template_skeleton.shape, dtype=template_skeleton.dtype
                )

            for flip_index, cur_theta in enumerate(thetas):

                score, synth_skel = centerline_accuracy(
                    theta=cur_theta,
                    template_skeleton=template_skeleton,
                    template_frame=template_frame,
                    template_measurements=template_measurements,
                    real_frame_orig=real_frame_orig,
                )

                all_scores[index][flip_index] = score
                all_skel[index][flip_index] = synth_skel

        # for one centerline case
        all_scores = all_scores.squeeze()
        all_skel = all_skel.squeeze()

        out_filename = os.path.join(temp_dir, f"compare_results_{chunk_index:09d}.pkl")

        with open(out_filename, "wb") as f:
            pickle.dump((all_scores, all_skel), f)

        results_queue.put(out_filename)



"""
Making images for the synthetic dataset
"""

# synthetic shape generator params:
# draw worm segments of a length 1/16 of the worm length
BODY_SEGMENTS_DIVIDER = 16
# draw patches a little larger than the worm thickness to add some background pixels
THICKNESS_MULTIPLIER = 1.2

# data augmentation parameters (applied when enable_random_augmentations is True):
# extra blurring happens 25% of the time
BLUR_PROBABILITY = 0.25
# draw tail on top of head 50% of the time
DRAW_TAIL_ON_TOP_PROBABILITY = 0.5
# extra translation max 5% of image size
MAX_OFFSET_PERCENT = 0.05
# when applying gaussian blur, minimum window size (must be an odd number)
GAUSSIAN_BLUR_MIN_SIZE = 3
# when applying gaussian blur, maximum window size (must be an odd number)
GAUSSIAN_BLUR_MAX_SIZE = 13
# when applying gaussian blur, maximum window size as a percentage of the image size
GAUSSIAN_BLUR_MAX_SIZE_PERCENT = 0.10
# change thickness multiplier by this amount + or -
THICKNESS_MULTIPLIER_OFFSET = 0.1
# multiply worm length by a random amount between 1 -/+ WORM_LENGTH_MULTIPLIER_OFFSET
WORM_LENGTH_MULTIPLIER_OFFSET = 0.1


from wormpose.images.synthetic.synthetic_dataset import SyntheticDataset



"""
Contains function to calculate distances between worm poses, either represented as angles or as skeletons
"""

import math

import numpy as np


def angle_distance(theta_a: np.ndarray, theta_b: np.ndarray) -> float:
    """
    Angle distance that takes into account the periodicity of angles
    """
    diff = np.abs(np.arctan2(np.sin(theta_a - theta_b), np.cos(theta_a - theta_b)))
    return diff.mean()


def _head_tail_diff(skel):
    return skel[-1][0] - skel[0][0], skel[-1][1] - skel[0][1]


def _cos_similarity(a, b):
    def _norm(x):
        return math.sqrt(x[0] * x[0] + x[1] * x[1])

    return (a[0] * b[0] + a[1] * b[1]) / (_norm(a) * _norm(b))


def skeleton_distance(skel_a: np.ndarray, skel_b: np.ndarray) -> float:
    """
    Cosine similarity between the two head to tail vectors of the input skeletons
    """
    return _cos_similarity(_head_tail_diff(skel_a), _head_tail_diff(skel_b))



"""
The Dataset loader: instantiates the FramesDataset, FeaturesDataset, FramePreprocessing, ResultsExporter (optional).
Also handles the resizing options.
"""

import os
from typing import List, Optional, Tuple

import pkg_resources

from wormpose.dataset.base_dataset import (
    BaseFramesDataset,
    BaseFramePreprocessing,
    BaseResultsExporter,
)
from wormpose.dataset.features import Features, calculate_crop_window_size, FeaturesDict
from wormpose.dataset.loaders.resizer import (
    frames_dataset_resizer,
    features_dataset_resizer,
    ResizeOptions,
)


class Dataset(object):
    def __init__(
        self,
        video_names: List[str],
        frames_dataset: BaseFramesDataset,
        features_dataset: FeaturesDict,
        frame_preprocessing: BaseFramePreprocessing,
        image_shape: Tuple[int, int],
        results_exporter: BaseResultsExporter,
    ):
        self.video_names = video_names
        self.frames_dataset = frames_dataset
        self.features_dataset = features_dataset
        self.frame_preprocessing = frame_preprocessing
        self.image_shape = image_shape
        self.results_exporter = results_exporter

    def num_frames(self, video_name):
        with self.frames_dataset.open(video_name) as frames:
            return len(frames)


def get_dataset_name(dataset_path: str) -> str:
    """
    Each dataset gets assigned a name: the root folder of the dataset

    :param dataset_path: Full path of the dataset
    :return: Name identifier for the dataset, simply use the basename of the path
        Each different dataset must have a unique basename in order to process several at once
    """
    return os.path.basename(os.path.normpath(dataset_path))


class _DummyResultsExporter(BaseResultsExporter):
    """
    Does nothing
    """

    def export(self, video_name: str, **kwargs):
        pass


def load_dataset(
    dataset_loader: str,
    dataset_path: str,
    selected_video_names: Optional[List[str]] = None,
    resize_options: ResizeOptions = None,
) -> Dataset:
    for entry_point in pkg_resources.iter_entry_points("worm_dataset_loaders"):
        if entry_point.name == dataset_loader:
            module = entry_point.load()

            frames_dataset_class = module.FramesDataset
            features_dataset_class = module.FeaturesDataset
            frame_preprocessing_class = module.FramePreprocessing

            frames_dataset, features_dataset, video_names = _load_dataset(
                frames_dataset_class, features_dataset_class, dataset_path, selected_video_names,
            )

            frame_preprocessing = frame_preprocessing_class()
            image_shape = calculate_crop_window_size(features_dataset)

            if resize_options is not None:
                resize_options.update_resize_factor(features_dataset)
                if resize_options.resize_factor != 1.0:
                    # reload frames and features dataset after resizing, also get new image_shape
                    frames_dataset_class = frames_dataset_resizer(
                        frames_dataset_class, resize_factor=resize_options.resize_factor
                    )
                    features_dataset_class = features_dataset_resizer(
                        features_dataset_class, resize_factor=resize_options.resize_factor,
                    )
                    frames_dataset, features_dataset, video_names = _load_dataset(
                        frames_dataset_class, features_dataset_class, dataset_path, selected_video_names,
                    )
                    image_shape = resize_options.get_image_shape(features_dataset)

            results_exporter = (
                module.ResultsExporter(dataset_path) if hasattr(module, "ResultsExporter") else _DummyResultsExporter()
            )

            return Dataset(
                video_names=video_names,
                features_dataset=features_dataset,
                frames_dataset=frames_dataset,
                frame_preprocessing=frame_preprocessing,
                image_shape=image_shape,
                results_exporter=results_exporter,
            )

    raise NotImplementedError(f"Dataset loader: '{dataset_loader}' not found in the package entry points.")


def _load_dataset(
    frames_dataset_class, features_dataset_class, dataset_path: str, selected_video_names,
):
    frames_dataset = frames_dataset_class(dataset_path)
    video_names = _resolve_video_names(frames_dataset, selected_video_names)

    raw_features_dataset = features_dataset_class(dataset_path, video_names)
    features_dataset = {}
    for video_name in video_names:
        features_dataset[video_name] = Features(raw_features_dataset.get_features(video_name))
    return frames_dataset, features_dataset, video_names


def _resolve_video_names(frames_dataset: BaseFramesDataset, selected_video_names: Optional[List[str]]):
    video_names = frames_dataset.video_names()

    if selected_video_names is not None:
        for video_name in selected_video_names:
            if video_name not in video_names:
                raise ValueError(f"Requested video '{video_name}' not found in dataset ({video_names})")
        video_names = selected_video_names

    _validate_video_names(video_names)

    return video_names


def _validate_video_names(video_names: List[str]):
    if len(video_names) == 0:
        raise ValueError(f"Video names list is empty, no video to analyze")
    if not all(isinstance(n, str) for n in video_names):
        raise NotImplementedError("Only video names as list of strings supported type.")
    if len(video_names) > len(set(video_names)):
        raise ValueError(f"Video names must be unique but duplicates were found: {video_names}.")



#!/usr/bin/env python

import setuptools

if __name__ == "__main__":
    with open("README.md") as f:
        long_description = f.read()

    version = {}
    with open("wormpose/version.py") as fp:
        exec(fp.read(), version)

    setuptools.setup(
        version=version["__version__"],
        long_description=long_description,
        long_description_content_type="text/markdown",
        include_package_data=True,
        package_data={"wormpose": ["resources/postures_model.json.gz"]},
        entry_points={
            "console_scripts": ["wormpose = wormpose.__main__:main"],
            "worm_dataset_loaders": [
                "tierpsy = wormpose.dataset.loaders.tierpsy",
                "sample_data = wormpose.dataset.loaders.sample_data",
            ],
        },
    )



#!/usr/bin/env python

"""
Post-processes WormPose results by interpolating over missing frames and smoothing
"""

import glob
import logging
import os
import tempfile
from argparse import Namespace
from typing import Sequence

import h5py
import numpy as np
import numpy.ma as ma
from scipy.interpolate import UnivariateSpline
from scipy.signal import savgol_filter

from wormpose.commands import _log_parameters
from wormpose.commands.utils.results_saver import ResultsSaver
from wormpose.commands.utils.time_sampling import resample_results
from wormpose.config import default_paths
from wormpose.config.default_paths import RESULTS_FILENAME, POSTPROCESSED_RESULTS_FILENAME, CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset.loader import get_dataset_name, Dataset
from wormpose.dataset.loader import load_dataset
from wormpose.images.scoring import ResultsScoring, ScoringDataManager
from wormpose.pose.eigenworms import load_eigenworms_matrix
from wormpose.pose.results_datatypes import BaseResults, OriginalResults

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _find_runs_boundaries(arr: np.ndarray) -> np.ndarray:
    """
    :return: list of pair of indexes [start, end] of the segments (runs of consecutive True values) boundaries
    """
    padded_arr = np.concatenate([[False], arr, [False]])
    return np.where(np.abs(np.diff(padded_arr)))[0].reshape(-1, 2)


def _get_valid_segments(is_valid_series: np.ndarray, max_gap_size: int, min_segment_size: int) -> Sequence:
    """
    :return: list of pair of indexes [start, end] of the valid segments in the data:
        They can contain small gaps <= max_gap_size but they need to be surrounded by valid data
        of length >= min_segment_size
    """

    # identify segments (consecutive runs of valid frames)
    segments_boundaries = _find_runs_boundaries(is_valid_series)

    # find the big segments of length greater or equal than min_segment_size
    big_segments_boundaries = [[start, end] for start, end in segments_boundaries if end - start >= min_segment_size]
    big_segments = np.full_like(is_valid_series, False)
    for gap_start, gap_end in big_segments_boundaries:
        big_segments[gap_start:gap_end] = True

    # identify gaps (consecutive runs of invalid frames between the big segments)
    gaps_boundaries = _find_runs_boundaries(~big_segments)

    # only keep the big gaps of length greater than max_gap_size
    big_gaps_boundaries = [[start, end] for start, end in gaps_boundaries if end - start > max_gap_size]
    big_gaps = np.full_like(is_valid_series, False)
    for gap_start, gap_end in big_gaps_boundaries:
        big_gaps[gap_start:gap_end] = True

    # the final valid segments are in-between the big gaps
    # they can contain small gaps but there are at least min_segment_size before and after the small gaps
    valid_segments_boundaries = _find_runs_boundaries(~big_gaps)

    return valid_segments_boundaries


class _SplineInterpolation(object):
    def __init__(self, frames_around: int = 3, spline_degree: int = 3):

        self.spline_degree = spline_degree
        self.frames_around = frames_around
        self.slope = np.linspace(0, 0.5, frames_around + 1)

    def interpolate_tseries(
        self, tseries: np.ndarray, segments_boundaries: Sequence, std_fraction: float
    ) -> np.ndarray:

        weight = 1 / (std_fraction * np.nanstd(tseries))

        tseries[~np.isnan(tseries)] = np.unwrap(tseries[~np.isnan(tseries)])
        new_tseries = np.full_like(tseries, np.nan)

        for t0, tf in segments_boundaries:
            new_tseries[t0:tf] = self._interpolate_segment(tseries[t0:tf], weight)

        return new_tseries

    def _interpolate_segment(self, tseries: np.ndarray, weight: float) -> np.ndarray:
        new_tseries = np.copy(tseries)

        nan_y = np.isnan(new_tseries)
        indices_nan = np.any(nan_y, axis=1)
        series_len = len(new_tseries)
        x = np.arange(series_len)
        new_tseries[nan_y] = 0.0

        w = self.build_weights(indices_nan, series_len, weight)

        # perform spline interpolation separately for each dimension
        for dim in range(new_tseries.shape[1]):
            y = new_tseries[:, dim]
            spl = UnivariateSpline(x, y, w=w, k=self.spline_degree, s=len(x))
            new_x = x
            new_weighted_y = spl(new_x)
            new_tseries[:, dim] = new_weighted_y

        return new_tseries

    def build_weights(self, indices_nan: np.ndarray, series_len: int, weight: float) -> np.ndarray:
        # setup weights: lower the weights closer to the edges
        w = np.full(series_len, weight)
        where_nan = np.where(indices_nan)[0]
        if len(where_nan) == 0:
            return w

        for idx in range(series_len):
            closest_nan_distance = np.min(np.abs(where_nan - idx))
            if closest_nan_distance <= self.frames_around:
                w[idx] = self.slope[closest_nan_distance] * weight

        return w


def _smooth_tseries(
    tseries: np.ndarray, smoothing_window_length: int, poly_order: int, segments_boundaries: Sequence,
) -> np.ndarray:
    if smoothing_window_length % 2 == 0:
        smoothing_window_length += 1  # need a odd number for smoothing_window_length

    new_tseries = np.full_like(tseries, np.nan)

    for t0, tf in segments_boundaries:
        if tf - t0 < smoothing_window_length or tf - t0 < poly_order:
            continue
        new_tseries[t0:tf] = savgol_filter(
            tseries[t0:tf], axis=0, window_length=smoothing_window_length, polyorder=poly_order,
        )

    return new_tseries


def _dorsal_ventral_flip_theta(theta: np.ndarray) -> np.ndarray:
    return 2 * np.pi - theta


def _thetas_to_modes(thetas: np.ndarray, eigenworms_matrix: np.ndarray) -> np.ndarray:
    return (thetas.T - thetas.mean(axis=1)).T.dot(eigenworms_matrix)


def _unwrap_ma(x: ma.MaskedArray):
    idx = ma.array(np.arange(0, x.shape[0]), mask=x.mask)
    idxc = idx.compressed()
    xc = x.compressed()
    dd = np.diff(xc)
    ddmod = np.mod(dd + np.pi, 2 * np.pi) - np.pi
    ddmod[(ddmod == -np.pi) & (dd > 0)] = np.pi
    phc_correct = ddmod - dd
    phc_correct[np.abs(dd) < np.pi] = 0
    ph_correct = np.zeros(x.shape)
    ph_correct[idxc[1:]] = phc_correct
    up = x + ph_correct.cumsum()
    return up


def _calculate_skeleton(theta: np.ndarray, args, dataset: Dataset, video_name: str) -> BaseResults:
    frames_timestamp = dataset.features_dataset[video_name].timestamp
    features = dataset.features_dataset[video_name]

    # resample time serie to have the same length as the number of frames
    theta_resampled = np.empty((dataset.num_frames(video_name),) + theta.shape[1:], dtype=theta.dtype)
    for cur_time, cur_theta in enumerate(theta):
        frame_index = np.where(frames_timestamp == cur_time)[0]
        theta_resampled[frame_index] = cur_theta

    results = BaseResults(theta=theta_resampled)
    ResultsScoring(
        frame_preprocessing=dataset.frame_preprocessing,
        num_process=args.num_process,
        temp_dir=args.temp_dir,
        image_shape=dataset.image_shape,
    )(
        results=results,
        scoring_data_manager=ScoringDataManager(
            video_name=video_name, frames_dataset=dataset.frames_dataset, features=features,
        ),
    )
    resample_results(results, features.timestamp)

    return results


def _parse_arguments(dataset_path: str, kwargs: dict):
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("max_gap_size") is None:
        kwargs["max_gap_size"] = 3
    if kwargs.get("min_segment_size") is None:
        kwargs["min_segment_size"] = 30
    if kwargs.get("smoothing_window") is None:
        kwargs["smoothing_window"] = 7
    if kwargs.get("poly_order") is None:
        kwargs["poly_order"] = 3
    if kwargs.get("std_fraction") is None:
        kwargs["std_fraction"] = 0.01
    if kwargs.get("eigenworms_matrix_path") is None:
        kwargs["eigenworms_matrix_path"] = None
    if kwargs.get("num_process") is None:
        kwargs["num_process"] = os.cpu_count()
    if kwargs.get("temp_dir") is None:
        kwargs["temp_dir"] = tempfile.gettempdir()
    kwargs["temp_dir"] = tempfile.mkdtemp(dir=kwargs["temp_dir"])

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def post_process(dataset_path: str, **kwargs):
    """
    Process the raw network results with interpolation and smoothing

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    results_root_dir = os.path.join(args.experiment_dir, default_paths.RESULTS_DIR)

    eigenworms_matrix = load_eigenworms_matrix(args.eigenworms_matrix_path)

    config = load_config(args.config)

    dataset = load_dataset(config.dataset_loader, dataset_path)

    spline_interpolation = _SplineInterpolation()

    results_files = list(sorted(glob.glob(os.path.join(results_root_dir, "*", RESULTS_FILENAME))))
    if len(results_files) == 0:
        raise FileNotFoundError("No results file to analyze was found")

    for results_file in results_files:
        video_name = os.path.basename(os.path.dirname(results_file))

        with h5py.File(results_file, "r") as results_f:

            try:
                results_raw = BaseResults(
                    theta=results_f["resolved"]["theta"][:],
                    skeletons=results_f["resolved"]["skeletons"][:],
                    scores=results_f["resolved"]["scores"][:],
                )
            except Exception:
                logger.error(f"Couldn't read results in file {results_file}.")
                continue

            results_orig = OriginalResults(
                theta=results_f["original"]["theta"][:], skeletons=results_f["original"]["skeletons"][:]
            )

            features = dataset.features_dataset[video_name]

            missing_values = np.any(np.isnan(results_raw.theta), axis=1)
            if missing_values.sum() == len(results_raw.theta):
                logger.warning(f"No valid result was found, stopping postprocessing for {video_name}")
                continue

            segments_boundaries = _get_valid_segments(
                is_valid_series=~missing_values, max_gap_size=args.max_gap_size, min_segment_size=args.min_segment_size,
            )
            # interpolate and smooth in angles space
            thetas_interp = spline_interpolation.interpolate_tseries(
                results_raw.theta, segments_boundaries, args.std_fraction
            )
            results_interp = _calculate_skeleton(thetas_interp, args, dataset, video_name)

            thetas_smooth = _smooth_tseries(thetas_interp, args.smoothing_window, args.poly_order, segments_boundaries,)
            results_smooth = _calculate_skeleton(thetas_smooth, args, dataset, video_name)

            flipped = False

            if features.ventral_side == "clockwise":
                results_orig.theta = _dorsal_ventral_flip_theta(results_orig.theta)
                results_raw.theta = _dorsal_ventral_flip_theta(results_raw.theta)
                results_interp.theta = _dorsal_ventral_flip_theta(results_interp.theta)
                results_smooth.theta = _dorsal_ventral_flip_theta(results_smooth.theta)
                flipped = True

            if eigenworms_matrix is not None:
                setattr(results_orig, "modes", _thetas_to_modes(results_orig.theta, eigenworms_matrix))
                setattr(results_raw, "modes", _thetas_to_modes(results_raw.theta, eigenworms_matrix))
                setattr(results_interp, "modes", _thetas_to_modes(results_interp.theta, eigenworms_matrix))
                setattr(results_smooth, "modes", _thetas_to_modes(results_smooth.theta, eigenworms_matrix))

        # save results
        results_saver = ResultsSaver(
            temp_dir=args.temp_dir, results_root_dir=results_root_dir, results_filename=POSTPROCESSED_RESULTS_FILENAME
        )

        metadata = {
            "max_gap_size": args.max_gap_size,
            "min_segment_size": args.min_segment_size,
            "smoothing_window": args.smoothing_window,
            "poly_order": args.poly_order,
            "std_fraction": args.std_fraction,
            "dorsal_ventral_flip": flipped,
        }

        results_saver.save(
            results={"orig": results_orig, "raw": results_raw, "interp": results_interp, "smooth": results_smooth},
            metadata=metadata,
            video_name=video_name,
        )
        logger.info(f"Post-processed worm: {video_name} {'(flipped dorsal-ventral)' if flipped else ''}")


def main():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("dataset_path", type=str)
    parser.add_argument(
        "--eigenworms_matrix_path", help="Path to optional eigenworms matrix to also save results in eigenworm modes",
    )
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    add_config_argument(parser)
    parser.add_argument(
        "--max_gap_size",
        type=float,
        help="Interpolate over missing values (gaps), as long"
        "as the consecutive length of the missing values is less than max_gap_size (frames)",
    )
    parser.add_argument(
        "--min_segment_size",
        type=float,
        help="Only segments of valid values of length greater than min_segment_size (frames)"
        "will be interpolated and smoothed",
    )
    parser.add_argument(
        "--std_fraction",
        type=int,
        help="The higher the guessed noise to signal ratio is, the smoother the interpolation will be",
    )
    parser.add_argument("--smoothing_window", type=int, help="smoothing window in frames")
    parser.add_argument("--poly_order", type=int, help="polynomial order in smoothing")
    parser.add_argument("--temp_dir", type=str, help="Where to store temporary intermediate results")
    parser.add_argument("--num_process", type=int, help="How many worker processes")
    args = parser.parse_args()

    post_process(**vars(args))


if __name__ == "__main__":
    main()



import json
import os

from typing import Tuple, List


def add_config_argument(parser):
    """
    For command line arguments, add the option to pass the path of the configuration file

    :param parser: Argparse parser
    """
    parser.add_argument(
        "--config",
        type=str,
        help="Path of the configuration file."
        " The file is created when generating the training dataset."
        " If not set, will look for it in the default location: at"
        " {work_dir}/{dataset_name}/config.json",
    )


class ExperimentConfig(object):
    """
    Data container for the experiment config, created when generating training data
    """

    def __init__(
        self,
        num_train_samples: int = None,
        num_eval_samples: int = None,
        image_shape: Tuple[int, int] = None,
        dataset_loader: str = None,
        theta_dimensions: int = None,
        resize_factor: float = None,
        video_names: List[str] = None,
    ):
        self.num_train_samples = num_train_samples
        self.num_eval_samples = num_eval_samples
        self.image_shape = image_shape
        self.dataset_loader = dataset_loader
        self.theta_dimensions = theta_dimensions
        self.resize_factor = resize_factor
        self.video_names = video_names


def save_config(experiment_config: ExperimentConfig, config_filepath: str):
    """
    Save the experiment config to a json file

    :param experiment_config: config object to save
    :param config_filepath: path where to write the config json file
    """
    with open(config_filepath, "w") as f:
        json.dump(experiment_config, f, indent=4, default=lambda x: x.__dict__)


def load_config(config_filepath: str) -> ExperimentConfig:
    """
    Load the experiment config from a json file

    :param config_filepath: path of the config json file to load
    :return: loaded config object
    """
    if not os.path.isfile(config_filepath):
        raise FileNotFoundError(f"Configuration file not found at path: '{config_filepath}'.")

    with open(config_filepath, "r") as f:
        return ExperimentConfig(**json.load(f))



"""
This module contains the WormPose API: abstract classes to subclass in order to add a custom dataset
"""

from abc import ABC, abstractmethod
from typing import List, Tuple
import numpy as np


class BaseFramePreprocessing(ABC):
    """
    Specific image processing logic to isolate the worm in the dataset images
    This object must be pickable (no inner functions for example)
    """

    @abstractmethod
    def process(self, frame: np.ndarray) -> Tuple[np.ndarray, int]:
        """
        Segment the worm object of interest in the image: returns a mask image of the same shape as frame,
        where the pixels belong to the worm object of interest are 1, and all the others are 0
        Also calculates the average value of the background pixels.

        :param frame: image to process
        :return: Segmentation mask image , background color
        """
        pass


class BaseFramesDataset(ABC):
    """
    Specific code to the dataset to access the frames only.
    A dataset is divided into several "videos". Each video contain a list of images or "frames".
    """

    @abstractmethod
    def video_names(self) -> List[str]:
        """
        A dataset is composed of several videos

        :return: A list of unique ids (string) identifying a video in the dataset
        """
        pass

    @abstractmethod
    def open(self, video_name: str):
        """
        The frames of the dataset are accessed trough a context manager object, in this way we have the option of not
        entirely loading a big image array in memory if possible

        :param video_name: One video unique id (should be one value of video_names())
        :return: A context manager object that can be used with the "with" python statement giving access to the
            frames (array of images) of the dataset
            Example use : frames = open("video0")
        """
        pass


class BaseFeaturesDataset(ABC):
    """
     Specific code to the dataset to access the features for each video of the dataset
     """

    @abstractmethod
    def get_features(self, video_name: str) -> dict:
        """
        Returns a dictionary of features

        :return: dictionary with keys: skeletons, head_width, midbody_width, tail_width, frame_rate, ventral_side, timestamp
            WHERE
            skeletons: Coordinates x y of the centerline for each frame in pixel coordinates,
            a numpy floating point array of shape (N number of frames, J number of joints, 2)
            The quality of the synthetic images will start degrading when J < 50, consider interpolating if less joints
            head_width: numpy floating point array of shape N
            midbody_width: numpy floating point array of shape N
            tail_width: numpy floating point array of shape N
            frame_rate: One float number for the frame rate of the video.
            ventral_side: Optional One string value for the entire video. \'clockwise\' or \'anticlockwise\'. If None, defaults to anticlockwise
            timestamp: Optional Timestamp of each frame, a numpy array of shape (N number of frames). If None, will consider each frame to be equidistant in time
        """
        pass


class BaseResultsExporter(ABC):
    """
    Optional Results Exporter
    """

    @abstractmethod
    def export(self, video_name: str, **kwargs):
        pass



"""
Simple FramesDataset implementation reading image files from a folder

Only one video is supported (one folder of images)
"""

import glob
import os
import cv2

from wormpose.dataset.base_dataset import BaseFramesDataset


class ImagesFolder(BaseFramesDataset):
    """
    Simple loading images in memory from one folder (one video_name)
    """

    class FramesReader(object):
        def __init__(self, filenames):
            self.data = [cv2.imread(f, cv2.IMREAD_GRAYSCALE) for f in filenames]

        def __enter__(self):
            return self.data

        def __exit__(self, exc_type, exc_value, traceback):
            pass

    def __init__(self, dataset_path: str, extension: str = "*.png"):
        self.filenames = sorted(glob.glob(os.path.join(dataset_path, extension)))
        if len(self.filenames) == 0:
            raise FileNotFoundError(f"Can't find {extension} images in folder {dataset_path}")

        self.video_name = os.path.basename(os.path.normpath(dataset_path))

    def video_names(self):
        return [self.video_name]

    def open(self, _):
        return self.FramesReader(self.filenames)



"""
This module contains the logic to resolve the head-tail orientation of a predicted video time series.
"""

import logging

import numpy as np
import numpy.ma as ma

from wormpose.pose.distance_metrics import angle_distance, skeleton_distance
from wormpose.pose.results_datatypes import (
    BaseResults,
    ShuffledResults,
    OriginalResults,
)

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# threshold to compare neighbor frames theta, to be considered continuous and belong to the same segment
CONTINUOUS_ANGLES_DIST_THRESHOLD = np.deg2rad(30)

# we consider frames to be part of the same segment if they are maximum this amount of seconds apart
# (and satisfy the distance threshold)
CONTINOUS_SEGMENT_TIME_WINDOW_SEC = 0.2

# discard too small segments less than this amount of seconds
MIN_SEGMENT_SIZE_SEC = 0.2

# don't align isolated segments that are more than this amount of seconds apart from aligned segments
MAXIMUM_GAP_ALLOWED_WITH_ADJACENT_SEGMENT_SEC = 1


def _init_partitioned_series(shuffled_series: np.ndarray):
    return ma.masked_all_like(shuffled_series)


def _set_partition(partitioned_series, shuffled_series, frame_index: int, partition: int):
    partitioned_series[frame_index][0] = shuffled_series[frame_index, partition]
    partitioned_series[frame_index][1] = shuffled_series[frame_index, 1 - partition]


class _PartitionedResults(BaseResults):
    def __init__(self, shuffled_results: ShuffledResults):

        self.cur_partition = -1
        self.partitions = ma.masked_all((len(shuffled_results),), dtype=int)
        self._shuffled_results = shuffled_results

        theta = _init_partitioned_series(shuffled_results.theta)
        skeletons = _init_partitioned_series(shuffled_results.skeletons)
        scores = _init_partitioned_series(shuffled_results.scores)
        super().__init__(theta=theta, skeletons=skeletons, scores=scores)

    def mask(self, indices):
        self.theta.mask[indices] = True
        self.skeletons.mask[indices] = True
        self.scores.mask[indices] = True
        self.partitions.mask[indices] = True

    def set_partition(self, frame_index: int, partition: int, new_partition: bool = False):
        if new_partition:
            self.cur_partition += 1

        _set_partition(self.theta, self._shuffled_results.theta, frame_index, partition)
        _set_partition(self.skeletons, self._shuffled_results.skeletons, frame_index, partition)
        _set_partition(self.scores, self._shuffled_results.scores, frame_index, partition)
        self.partitions[frame_index] = self.cur_partition

    def _get_partition_indices(self, partition_index: int):
        return np.where(self.partitions == partition_index)[0]

    def get_segments(self):
        all_partitions_indexes = np.unique(self.partitions.filled(-1))
        return [
            self._get_partition_indices(partition_index)
            for partition_index in all_partitions_indexes
            if partition_index >= 0
        ]


class _ResolvedResults(BaseResults):
    def __init__(self, partitioned_results: _PartitionedResults):
        self._partitioned_results = partitioned_results
        theta = _init_unified_series(partitioned_results.theta)
        skeletons = _init_unified_series(partitioned_results.skeletons)
        scores = _init_unified_series(partitioned_results.scores)
        super().__init__(theta=theta, skeletons=skeletons, scores=scores)

    def resolve(self, segment, segment_alignment):
        self.scores[segment] = self._partitioned_results.scores[segment][:, segment_alignment]
        self.skeletons[segment] = self._partitioned_results.skeletons[segment][:, segment_alignment]
        self.theta[segment] = self._partitioned_results.theta[segment][:, segment_alignment]

    def mask(self, indices):
        self.theta.mask[indices] = True
        self.skeletons.mask[indices] = True
        self.scores.mask[indices] = True

    def num_valid(self):
        return np.sum(~self.scores.mask)


class _FinalResults(BaseResults):
    @classmethod
    def from_resolved(cls, resolved_results: _ResolvedResults):
        return _FinalResults(
            theta=resolved_results.theta.filled(np.nan),
            skeletons=resolved_results.skeletons.filled(np.nan),
            scores=resolved_results.scores.filled(np.nan),
        )

    @classmethod
    def from_shuffled(cls, shuffled_results: ShuffledResults):
        return _FinalResults(
            theta=np.full_like(shuffled_results.theta[:, 0], np.nan),
            skeletons=np.full_like(shuffled_results.scores[:, 0], np.nan),
            scores=np.full_like(shuffled_results.skeletons[:, 0], np.nan),
        )


def _make_continuous_partitions(
    shuffled_results: ShuffledResults, score_threshold: float, frame_rate: float
) -> _PartitionedResults:
    time_window = max(1, int(frame_rate * CONTINOUS_SEGMENT_TIME_WINDOW_SEC))
    min_segment_size = max(1, int(frame_rate * MIN_SEGMENT_SIZE_SEC))

    partitioned_results = _PartitionedResults(shuffled_results)

    # discard low score frames early (use the maximum value of both scores for now)
    good_score_frames = np.where(ma.greater_equal(ma.max(shuffled_results.scores, axis=1), score_threshold))[0]

    for frame_index in good_score_frames:

        prev_theta = partitioned_results.theta[frame_index - min(time_window, frame_index) : frame_index, 0]

        # if there is a big gap > time_window we start a new partition, with a random value (0)
        if np.all(np.any(prev_theta.mask, axis=1)):
            partitioned_results.set_partition(frame_index=frame_index, partition=0, new_partition=True)
        # otherwise we look in the time_window close past the closest non nan frame see if we can continue the
        # partition as long as the values stay continuous
        else:
            last_valid_index = np.where(~np.any(prev_theta.mask, axis=1))[0][-1]
            dists = [
                angle_distance(shuffled_results.theta[frame_index, k, :], prev_theta[last_valid_index],)
                for k in range(2)
            ]
            partition = int(np.argmin(dists))
            if dists[partition] < CONTINUOUS_ANGLES_DIST_THRESHOLD:
                partitioned_results.set_partition(frame_index=frame_index, partition=partition)

    # discard short segments
    for cur_partition_indices in partitioned_results.get_segments():
        if len(cur_partition_indices) < min_segment_size:
            partitioned_results.mask(cur_partition_indices)

    return partitioned_results


def _align_segments_with_labels(segments, partitioned_skeletons, labelled_skeletons, min_labelled=5):
    """
    Match the head/tail alignment with the results of the classical tracking in each of the segments,
     if there is enough labelled data in the segment
    """
    segments_alignment = ma.masked_all((len(segments),), dtype=np.uint8)
    for segment_index, segment in enumerate(segments):
        segment_skeletons = labelled_skeletons[segment]
        non_nan_labelled = np.any(~np.isnan(segment_skeletons), axis=(1, 2))
        labels_count = np.sum(non_nan_labelled)
        non_masked = ~np.any(partitioned_skeletons[segment].mask, axis=(1, 2, 3))
        to_compare = np.logical_and(non_nan_labelled, non_masked)

        similarity_scores = []
        for label_skel, partitioned_skeleton in zip(
            segment_skeletons[to_compare], partitioned_skeletons[segment][to_compare]
        ):
            dists = [skeleton_distance(label_skel, x) for x in partitioned_skeleton]
            similarity_scores.append(dists)

        if len(similarity_scores) > 0:
            mean_similarity_scores = np.mean(similarity_scores, axis=0)
            if mean_similarity_scores[0] * mean_similarity_scores[1] < 0 and labels_count > min_labelled:
                segments_alignment[segment_index] = np.argmax(mean_similarity_scores)

    return segments_alignment


def _calculate_smallest_gap_to_adjacent(segment_index, segments, segments_alignment):
    # evaluate how far away this segment is from known values
    score = np.nan
    segment_offset = np.nan
    if segment_index - 1 >= 0 and not segments_alignment.mask[segment_index - 1]:
        gap = segments[segment_index][0] - segments[segment_index - 1][-1]
        score = gap
        segment_offset = -1
    if segment_index + 1 < len(segments_alignment) and not segments_alignment.mask[segment_index + 1]:
        gap = segments[segment_index + 1][0] - segments[segment_index][-1]
        if np.isnan(score) or gap < score:
            score = gap
            segment_offset = 1

    return score, segment_offset


def _align_unlabelled_segments_with_adjacents(segments, segments_alignment, partitioned_skeletons, frame_rate: float):
    """
    Resolve the unaligned segments by comparing with adjacent segments,
    starting with the segments that have the least frames gap between an adjacent trusted segment
    Don't align isolated segments which a big gap between trusted segments
    """
    maximum_gap_allowed = max(1, int(frame_rate * MAXIMUM_GAP_ALLOWED_WITH_ADJACENT_SEGMENT_SEC))
    # ensure that if no segments have been aligned at all, pick one solution randomly to start
    if np.all(segments_alignment.mask):
        logger.info("There are no trusted segments with head decision to resolve the whole video, stopping analysis.")
        return segments_alignment

    # fix in priority the segments with known adjacent frames with little gap
    # until all segments are aligned except the isolated ones (further than maximum_gap_allowed)
    unaligned = np.where(segments_alignment.mask)[0]
    while len(unaligned) > 0:
        # we first pick the best candidate segment to align (there are known frames nearby before or after or both)
        all_gaps = [
            _calculate_smallest_gap_to_adjacent(
                segment_index=x, segments=segments, segments_alignment=segments_alignment,
            )
            for x in unaligned
        ]
        segment_to_fix_index = np.nanargmin(all_gaps, axis=0)[0]
        gap_to_adjacent_segment, adjacent_segment_offset = all_gaps[segment_to_fix_index]

        # abort if only isolated segments are left
        if gap_to_adjacent_segment > maximum_gap_allowed:
            break

        cur_segment_index = unaligned[segment_to_fix_index]
        cur_segment_skeleton = partitioned_skeletons[segments[cur_segment_index]]

        adjacent_segment_index = cur_segment_index + adjacent_segment_offset
        adjacent_alignment = segments_alignment[adjacent_segment_index]
        adjacent_segment = segments[adjacent_segment_index]
        adjacent_segment_skeleton = partitioned_skeletons[adjacent_segment][:, adjacent_alignment]

        if adjacent_segment_offset == -1:
            closest_unaligned_skeleton = cur_segment_skeleton[0]  # first frame of cur segment
            closest_known_skeleton = adjacent_segment_skeleton[-1]  # last frame of prev segment
        elif adjacent_segment_offset == 1:
            closest_unaligned_skeleton = cur_segment_skeleton[-1]  # last frame of cur segment
            closest_known_skeleton = adjacent_segment_skeleton[0]  # first frame of next segment
        else:
            raise ValueError()

        dists = [skeleton_distance(closest_known_skeleton, skel) for skel in closest_unaligned_skeleton]
        segments_alignment[cur_segment_index] = int(np.argmax(dists))

        unaligned = np.where(segments_alignment.mask)[0]

    return segments_alignment


def _init_unified_series(mixed_series):
    return ma.masked_all((mixed_series.shape[0],) + mixed_series.shape[2:], dtype=mixed_series.dtype)


def resolve_head_tail(
    shuffled_results: ShuffledResults, original_results: OriginalResults, frame_rate: float, score_threshold,
) -> BaseResults:
    len_series = len(shuffled_results)

    # Create continuous segments without jumps
    partitioned_results = _make_continuous_partitions(
        score_threshold=score_threshold, frame_rate=frame_rate, shuffled_results=shuffled_results,
    )
    segments = partitioned_results.get_segments()

    if len(segments) == 0:
        logger.error(
            f"Couldn't find any continuous segments of predicted data above the threshold {score_threshold},"
            f" stopping analysis."
        )
        return _FinalResults.from_shuffled(shuffled_results)

    # Choose each segment global alignment by comparing with labelled data
    segments_alignment = _align_segments_with_labels(
        segments, partitioned_results.skeletons, original_results.skeletons
    )

    # Fix unaligned segments here by comparing skeletons with neighboring segments iteratively
    segments_alignment = _align_unlabelled_segments_with_adjacents(
        segments, segments_alignment, partitioned_results.skeletons, frame_rate
    )

    # Compile results
    resolved_results = _ResolvedResults(partitioned_results)
    for segment, segment_alignment in zip(segments, segments_alignment):
        if not ma.is_masked(segment_alignment):
            resolved_results.resolve(segment, segment_alignment)

    # Filter the final results again by score threshold
    low_scores_indices = np.where(ma.masked_less(resolved_results.scores, score_threshold).mask)[0]
    resolved_results.mask(low_scores_indices)

    num_success = resolved_results.num_valid()
    original_num_success = np.any(~np.isnan(original_results.skeletons), axis=(1, 2)).sum()
    logger.info(
        f"Resolved head/tail, {num_success} out of {len_series} frames analyzed successfully "
        f"({float(num_success) / len_series * 100:.1f}%) (original features : {original_num_success}"
        f" or {(float(original_num_success) / len_series * 100):.1f}% of total)"
    )
    if num_success < original_num_success:
        logger.warning(f"Original results had {original_num_success - num_success} more successfully analyzed frames!")

    return _FinalResults.from_resolved(resolved_results)



"""
Simple BaseFramePreprocesing implementation
"""

from typing import Optional, Tuple, Callable

import cv2
import numpy as np

from wormpose.dataset.base_dataset import BaseFramePreprocessing
from wormpose.dataset.image_processing.image_utils import segment_foreground, OtsuThreshold


class SimpleFramePreprocessing(BaseFramePreprocessing):
    def __init__(
        self,
        foreground_dilate_struct_element=cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5)),
        foreground_close_struct_element=cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3)),
        threshold_fn: Callable = OtsuThreshold(blur_kernel=(5, 5)),
    ):
        self.foreground_dilate_struct_element = foreground_dilate_struct_element
        self.foreground_close_struct_element = foreground_close_struct_element
        self.threshold_fn = threshold_fn

    def process(self, frame: np.ndarray, background_threshold: Optional[int] = None) -> Tuple[np.ndarray, int]:
        return segment_foreground(
            frame, self.foreground_close_struct_element, self.foreground_dilate_struct_element, self.threshold_fn,
        )



from typing import Callable


def _log_parameters(logger: Callable, params: dict):
    """
    Utility function to log script parameters with a more readable formatting
    """
    logger("\n\t" + ", \n\t".join([f"{x[0]}: {x[1]}" for x in params.items()]))


from wormpose.commands.calibrate_dataset import calibrate
from wormpose.commands.evaluate_model import evaluate
from wormpose.commands.export_results import export
from wormpose.commands.generate_training_data import generate
from wormpose.commands.postprocess_results import post_process
from wormpose.commands.predict_dataset import predict
from wormpose.commands.train_model import train
from wormpose.commands.visualize_results import visualize
